"""
Coordinate the occurrence of events.
"""
import itertools
import collections
import heapq

from . import system

class Harmony(object):
	"""
	All purpose event scheduler.

	Coorindate the production of events according to a time delay against a monotonic timer.

	Each Defers instance manages the production of a set of events according to the
	prescribed delay. When an event is scheduled, the amount of time specified is added to the
	meter's current position. When the meter surpasses that value, the event is emitted.

	Events are arbitrary Python objects. Often, they will be tasks to be performed.

	Harmony differs from Python's scheduler as there is no attempt to manage the sleep
	period. Rather, this responsibility is offloaded onto the user in order to keep
	functionality isolated.
	"""
	unit = 'nanosecond'

	def __init__(self,
		Chronometer = system.Chronometer,
		Identifiers = itertools.count,
		DefaultDict = collections.defaultdict,
		Sequence = collections.deque
	):
		self.meter = Chronometer()
		self.identifiers = iter(Identifiers())
		self.heap = []
		# RPiT to (Eventi) Id's Mapping
		self.schedule = DefaultDict(Sequence)
		# (Event) Id to Event Mapping
		self.events = {}
		self.cancellations = set()

	def period(self):
		"""
		The period before the next event should occur.

		When combining Harmony instances, this method can be used to identify
		when this Harmony instance should be processed.
		"""
		try:
			smallest = self.heap[0]
			return smallest.__class__(smallest - self.meter.snapshot())
		except IndexError:
			return None

	def allocate(self):
		# Internal
		for ident in self.identifiers:
			if ident not in self.events:
				return ident

	def cancel(self, *ident):
		"""
		Cancel the scheduled events with the given identifiers.
		"""
		# Update the set of cancellations immediately.
		# This set is used as a filter by Defer cycles.
		self.cancellations.update(ident)

	def put(self, *schedules, push = heapq.heappush):
		"""
		put(*schedules)

		:param schedules: The Measure, Event pairs.
		:type schedules: (:py:class:`rhythm.lib.Measure`, :py:class:`object`)
		:returns: A sequence of event identifiers that can be used for cancellation.
		:rtype: [:py:class:`int`]

		Scheduls the given events.
		"""
		snapshot = self.meter.snapshot()
		events = []
		try:
			curmin = self.heap[0]
		except:
			curmin = None

		for assignment in schedules:
			measure, event = assignment
			eid = self.allocate()
			pit = measure.__class__(measure + snapshot)
			events.append(eid)
			try:
				self.events[eid] = event
				self.schedule[pit].append(eid)
				push(self.heap, pit)
			except:
				self.cancel(eid)
				raise

		return events

	def get(self, pop = heapq.heappop, push = heapq.heappush):
		"""
		get()

		:returns: The events whose scheduled point in time has elapsed.
		:rtype: [(:py:class:`rhythm.lib.Measure`, :py:class:`object`)]

		Return all events whose sheduled delay has elapsed according to the Chronometer.

		The pairs within the returned sequence consist of a Measure and the Event. The
		measure is the amount of time that has elapsed since the scheduled time.
		"""
		events = []
		cur = self.meter.snapshot()
		while self.heap:
			# repeat some work in case of concurrent pop
			item = pop(self.heap)
			overflow = item.__class__(cur - item)

			# the set of callbacks have passed their time.
			if overflow < 0:
				# put it back...
				push(self.heap, item)
				break
			else:
				eids = self.schedule[item]
				while eids:
					x = eids.popleft()
					if x in self.cancellations:
						# filter any cancellations
						# schedule is already popped, so remove event and cancel*
						self.events.pop(x, None)
						self.cancellations.discard(x)
					else:
						if x in self.events:
							events.append((overflow, self.events.pop(x)))
				if not eids:
					self.schedule.pop(item, None)
		return events
